// http://www.hugi.scene.org/online/coding/hugi%2023%20-%20bzminima.htm

#include "malloc.h"

// DOES NOT CONVERGE !

/* A generic polynomial solver using Newton's method, with
   the given tolerance.  Assumes that the n roots are all real. */
void polysolve (float *f, float *x, int n, float tol) {
  float oldx, *b, y, d, *tmp;
  int i;
  b = alloca(sizeof(float) * (n + 1));
  do {
    *x = 0.0;
    do {
      oldx = *x;
      /* Divide the polynomial by x-x0 in b(n), get the value in x=x0
         in y and the derivative in d. This is called the nested
         multiplication algorithm and recalls high school's Ruffini's
         rule. */
      b[n] = 0.0;
      b[n - 1] = d = f[n];
      for (i = n - 1; i > 0; i--) {
        b[i - 1] = *x * b[i] + f[i];
        d = *x * d + b[i - 1];
      }
      y = *x * b[0] + f[0];
      /* We must not divide by zero even if x got exactly on a maximum
         or minimum */
      if (fabs(d) < 1.0e-10)
        *x += tol * 2.0;
      else
        *x -= y / d;
    } while (fabs(oldx - *x) >= tol);
    /* The results of division by (x - *x) are already in b; this
       is the polynomial for which we are going to find the next
       root. */
    tmp = f;
    f = b;
    b = tmp;
    x++;
  } while(n-- > 0);
  xfree(b);
}

void chebyshevt (float *x, int n, float a, float b) {
  float *f[3], *g;
  int i, j;
  f[0] = malloc(sizeof(float) * (n + 1));
  f[1] = malloc(sizeof(float) * (n + 1));
  f[2] = malloc(sizeof(float) * (n + 1));
  /* Initialize the polynomials */
  for (i = 0; i < n; i++)
    f[0][i] = f[1][i] = f[2][i] = 0.0;
  f[0][0] = f[1][1] = 1.0;
  for (i = 2; i <= n; i++) {
    /* Compute the recurrence relationship */
    f[2][0] = 0.0;
    for (j = 0; j < n; j++) {
      f[2][j + 1] = 2.0 * f[1][j];
      f[2][j] -= f[0][j];
    }
    /* Rotate the arrays */
    g = f[0];
    f[0] = f[1];
    f[1] = f[2];
    f[2] = g;
  }
  /* Now the polynomial is in f[1] */
  polysolve(f[1], x, n, 1.0e-10);
  for (i = 0; i < n; i++)
    x[i] = (a + b)/2 + (b - a)/2 * x[i];
  xfree(f[0]); xfree(f[1]); xfree(f[2]);
}

static int calc_chebyshevt (lua_State *L) {
  float *x, a, b;
  size_t i, n;
  n = agn_checkposint(L, 1);
  a = agn_checknumber(L, 2);
  b = agn_checknumber(L, 3);
  x = malloc(sizeof(float)*n);
  for (i=0; i < n; i++) x[i] = i+1;
  chebyshevt(x, n, a, b);  /* e.g. x, 5, .25, 4 */
  agn_createreg(L, n);
  for (i=0; i < n; i++) {
    lua_regsetinumber(L, -1, i + 1, x[i]);
  }
  xfree(x);
  return 1;
}


  {"chebyshevt", calc_chebyshevt},