#! /bin/sh
# Compile and install skript for the plus package for Solaris 10 (Sparc and x86)
#
# Execute this batch file in the .../agena/src folder by typing:
# sh makeplussolaris.sh
#
# Note that libgcc_s.so and libgcc_s.so.1 must exist in /usr/lib and
# /usr/local/lib so that the modules can be used.

# delete *.o files not deleted by make clean
for i in ads.o calc.o linalg.o stats.o fractals.o phq.o xbase.o net.o bags.o astro.o moon.o sunriset.o skycrane.o llist.o hashes.o numarray.o fastmath.o bytes.o sema.o semalib.o dual.o utf8.o mp.o mpf.o luasys.o charbuf.o com.o heaps.o long.o
do
   if [ -f i ]; then
      rm i
   fi
done

OPTIONS="-O3 -fpic -Wno-attributes -fgnu89-inline"
OPTFINAL="-O3 -shared -fpic -Wno-attributes -fgnu89-inline"

printf "Compiling ADS ... "
gcc $OPTIONS -c -o ads.o ads.c
# we need to compile and link against vector in Sun Solaris
gcc $OPTIONS -c -o vecoff64.o vecoff64.c
gcc $OPTFINAL -o ads.so ads.o vecoff64.o
printf "done.\n"

if [ -f ../phq/phq.c ]; then
   printf "Compiling phonetiQs ... "
   gcc $OPTIONS -c -o phq.o ../phq/phq.c
   gcc $OPTFINAL -o phq.so phq.o
   printf "done.\n"
   strip phq.so
   mv phq.so ../phq
fi

printf "Compiling calc ... "
gcc $OPTIONS -c -o calc.o calc.c
gcc $OPTIONS -c -o interp.o interp.c
gcc $OPTIONS -c -o rlhmath.o rlhmath.c
gcc $OPTFINAL -o calc.so calc.o interp.o rlhmath.o
printf "done.\n"

printf "Compiling stats ... "
gcc $OPTIONS -c -o stats.o stats.c
gcc $OPTFINAL -o stats.so stats.o
printf "done.\n"

printf "Compiling linalg ... "
gcc $OPTIONS -c -o linalg.o linalg.c
gcc $OPTFINAL -o linalg.so linalg.o
printf "done.\n"

printf "Compiling fractals ... "
gcc $OPTIONS -c -o fractals.o fractals.c
gcc $OPTFINAL -o fractals.so fractals.o
printf "done.\n"

printf "Compiling astro ... "
gcc $OPTIONS -c -o astro.o astro.c
gcc $OPTIONS -c -o moon.o moon.c
gcc $OPTIONS -c -o sunriset.o sunriset.c
gcc $OPTFINAL -o astro.so astro.o sunriset.o moon.o
printf "done.\n"

printf "Compiling xbase ... "
gcc $OPTIONS -c -o xbase.o xbase.c
gcc $OPTFINAL -o xbase.so xbase.o
printf "done.\n"

printf "Compiling net ... "
gcc $OPTIONS -c -o net.o net.c
gcc $OPTFINAL -o net.so net.o -lsocket -lnsl
printf "done.\n"

printf "Compiling bags ... "
gcc $OPTIONS -c -o bags.o bags.c
gcc $OPTFINAL -o bags.so bags.o
printf "done.\n"

printf "Compiling llist ... "
gcc $OPTIONS -c -o llist.o llist.c
gcc $OPTFINAL -o llist.so llist.o
printf "done.\n"

printf "Compiling hashes ... "
gcc $OPTIONS -c -o hashes.o hashes.c
gcc $OPTFINAL -o hashes.so hashes.o
printf "done.\n"

printf "Compiling skycrane ... "
gcc $OPTIONS -c -o skycrane.o skycrane.c
gcc $OPTFINAL -o skycrane.so skycrane.o
printf "done.\n"

printf "Compiling numarray ... "
gcc $OPTIONS -c -o numarray.o numarray.c
gcc $OPTFINAL -o numarray.so numarray.o
printf "done.\n"

printf "Compiling zx ... "
gcc $OPTIONS -c -o zx.o zx.c
gcc $OPTFINAL -o zx.so zx.o
printf "done.\n"

printf "Compiling bloom ... "
gcc $OPTIONS -c -o bloom.o bloom.c
gcc $OPTFINAL -o bloom.so bloom.o
printf "done.\n"

printf "Compiling factory ... "
gcc $OPTIONS -c -o factory.o factory.c
gcc $OPTFINAL -o factory.so factory.o
printf "done.\n"

printf "Compiling fastmath ... "
gcc $OPTIONS -Wno-strict-aliasing -c -o fastmath.o fastmath.c
gcc $OPTFINAL -Wno-strict-aliasing -o fastmath.so fastmath.o
printf "done.\n"

printf "Compiling sema ... "
gcc $OPTIONS -c -o semalib.o semalib.c
gcc $OPTIONS -c -o sema.o sema.c
gcc $OPTFINAL -o sema.so sema.o semalib.o
printf "done.\n"

printf "Compiling dual ... "
gcc $OPTIONS -c -o dual.o dual.c
gcc $OPTFINAL -o dual.so dual.o
printf "done.\n"

printf "Compiling utf8 ... "
gcc $OPTIONS -std=c99 -c -o utf8.o utf8.c
gcc $OPTFINAL -std=c99 -o utf8.so utf8.o
printf "done.\n"

printf "Compiling aconv ... "
gcc $OPTIONS -std=c99 -liconv -c -o aconv.o aconv.c
gcc $OPTFINAL -std=c99 -liconv -o aconv.so aconv.o
printf "done.\n"

printf "Compiling mp ... "
gcc $OPTIONS -std=c99 -lgmp -c -o mp.o mp.c
gcc $OPTFINAL -std=c99 -lgmp -o mp.so mp.o
printf "done.\n"

printf "Compiling mpf ... "
gcc $OPTIONS -std=c99 -lmpfr -lgmp -c -o mpf.o mpf.c
gcc $OPTFINAL -std=c99 -lmpfr -lgmp -o mpf.so mpf.o
printf "done.\n"

printf "Compiling com ... "
gcc $OPTIONS -DPLUS -c -o charbuf.o charbuf.c
gcc $OPTIONS -c -o luasys.o luasys.c
gcc $OPTIONS -c -o com.o com.c
gcc $OPTFINAL -o com.so com.o charbuf.o luasys.o
printf "done.\n"

printf "Compiling bytes ... "
gcc $OPTIONS -c -o bytes.o bytes.c
gcc $OPTFINAL -o bytes.so bytes.o charbuf.o
printf "done.\n"

printf "Compiling heaps ... "
gcc $OPTIONS -c -o heaps.o heaps.c
gcc $OPTFINAL -o heaps.so heaps.o
printf "done.\n"

printf "Compiling long ... "
gcc $OPTIONS -c -o long.o long.c
gcc $OPTFINAL -o long.so long.o
printf "done.\n"

#printf "Compiling func ... "
#gcc $OPTIONS -c -o func.o func.c
#gcc $OPTFINAL -o func.so func.o
#printf "done.\n"

printf "Slimming and moving packages ... "
strip *.so
mv ads.so ../lib
mv calc.so ../lib
mv stats.so ../lib
mv linalg.so ../lib
mv fractals.so ../lib
mv xbase.so ../lib
mv net.so ../lib
mv bags.so ../lib
mv llist.so ../lib
mv hashes.so ../lib
mv astro.so ../lib
mv skycrane.so ../lib
mv numarray.so ../lib
mv zx.so ../lib
mv bloom.so ../lib
mv factory.so ../lib
mv fastmath.so ../lib
mv bytes.so ../lib
mv sema.so ../lib
mv dual.so ../lib
mv utf8.so ../lib
mv aconv.so ../lib
mv mp.so ../lib
mv com.so ../lib
mv heaps.so ../lib
mv long.so ../lib
#mv func.so ../lib

printf "done.\n"

echo Installed all libraries into /lib folder.
echo All done.

