#ifndef semalib_h
#define semalib_h

#include <math.h>
#include <stdint.h>  /* for UINT32_MAX */


typedef struct Sema {
  uint32_t *s;  /* semaphore */
  uint32_t n;   /* total number of uint32_t chunks */
  uint32_t ff;  /* first chunk with a zero bit */
  uint64_t c;   /* current allocated semaphore, estimate only for sema.close may `corrupt` the value. */
  char     scon;  /* is simple counter currently in use ? 1 = yes, 0 = no */
  int64_t  sc;
} Sema;


#define SLOTS				8
#define SIZESLOT            (sizeof(uint32_t))                        /* usually 4 */
#define SIZECHUNK			(SLOTS*SIZESLOT)                          /* usually 32 = 4 * 8 at package initialisation */
#define SETBIT(x,pos,bit)	((x) & ~(1 << (pos))) | ((bit) << (pos))  /* pos counting from 0 */
#define GETBIT(x,pos)		((x & (1 << (pos))) != 0)                 /* pos counting from 0 */

/* See also: https://stackoverflow.com/questions/31393100/how-to-get-position-of-right-most-set-bit-in-c */
#define FIRSTZEROBIT(x)		(tools_uintlog2(~(x) & ((x) + 1)))        /* rightmost unset bit, return counting from 0 */
#define FIRSTONEBIT(x)		(tools_uintlog2((x) & ((~x) + 1)))        /* rightmost set bit, return counting from 0 */
#define LASTONEBIT(x)		(tools_msb((x)) - 1)                      /* leftmost set bit, return counting from 0 */

#undef AGN_LASTCONTINT
#define AGN_LASTCONTINT     9007199254740992ULL                       /* = (pow(2, 53)), 2.10.0/2.17.1 change, may not be 16-bit compliant FIXME */

Sema *semalib_new (void);
int  semalib_init (Sema *s);
void semalib_destroy (Sema *s);
int  semalib_expand (Sema *s, uint32_t nslots);
int  semalib_open (Sema *s, uint32_t semano);
int  semalib_close (Sema *s, uint32_t semano);
int  semalib_isopen (Sema *s, uint32_t semano);
int  semalib_reset (Sema *s);
int  semalib_shrink (Sema *s);

#endif