/*
** $Id: strarray.c,v 0.1 15.10.2020 $
** strarray - String Array Library
** See Copyright Notice in agena.h
*/

#define strarray_c
#define LUA_LIB

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <unistd.h>

#include "agena.h"
#include "agnxlib.h"
#include "agenalib.h"
#include "lstate.h"
#include "agnhlps.h"

/* for strarray.used: */
#include "lstring.h"
#include "lapi.h"
#include "strarray.h"


#if !(defined(LUA_DOS) || defined(__OS2__) || defined(LUA_ANSI))
#define AGENA_STRARRAYLIBNAME "strarray"
LUALIB_API int (luaopen_strarray) (lua_State *L);
#endif


static const struct luaL_Reg strarray_arraylib [] = {  /* metamethods for numeric userdata `n` */
  {"__eq",       mt_eeq},              /* equality mt */
  {"__eeq",      mt_eeq},              /* strict equality mt */
  {"__index",    strarray_subarray},   /* n[p], with p the index, counting from 1, NO support of n[p to q] indexing since the VM can only return one value */
  {"__writeindex", strarray_setitem},  /* n[p] := value, with p the index, counting from 1 */
  {"__tostring", mt_array2string},     /* for output at the console, e.g. print(n) */
  {"__size",     mt_getsize},          /* metamethod for `size` operator */
  {"__empty",    mt_empty},            /* metamethod for `empty` operator */
  {"__filled",   mt_filled},           /* metamethod for `filled` operator */
  {"__gc",       mt_arraygc},          /* please do not forget garbage collection */
  {"__in",       mt_in},               /* metamethod for `in` operator */
  {"__notin",    mt_notin},            /* metamethod for `notin` operator */
  {NULL, NULL}
};


static const luaL_Reg strarraylib[] = {
  {"convert",    strarray_convert},
  {"cycle",      strarray_cycle},
  {"double",     strarray_double},
  {"getitem",    strarray_getitem},
  {"getbit",     strarray_getbit},
  {"getsize",    mt_getsize},
  {"include",    strarray_include},
  {"iterate",    strarray_iterate},
  {"map",        strarray_map},
  {"purge",      strarray_purge},
  {"read",       strarray_read},
  {"resize",     strarray_resize},
  {"satisfy",    strarray_satisfy},
  {"setbit",     strarray_setbit},
  {"setitem",    strarray_setitem},
  {"sort",       strarray_sort},
  {"toarray",    strarray_toarray},
  {"toreg",      strarray_toreg},
  {"toseq",      strarray_toseq},
  {"used",       strarray_used},
  {"whereis",    strarray_whereis},
  {"write",      strarray_write},
  {NULL, NULL}
};


/*
** Open strarray library
*/
LUALIB_API int luaopen_strarray (lua_State *L) {
  luaL_newmetatable(L, "strarray");  /* metatable for strarray userdata, adds it to the registry with key 'strarray' */
  luaL_register(L, NULL, strarray_arraylib);  /* assign C metamethods to this metatable */
  luaL_register(L, AGENA_STRARRAYLIBNAME, strarraylib);
  return 1;
}
